/*
  $Header: /u/spert/src/rprf/RCS/rprf.h,v 1.14 1996/11/01 06:06:23 krste Exp $

  Routines to help with performance testing.
*/

#ifndef rprf_h_
#define rprf_h_

#ifdef __cplusplus
extern "C" {
#endif

#include <assert.h>

#include "IPM_timer.h"

extern long rprf_n_inner;
extern long rprf_n_outer;

/* Types of raw time reporting. */
enum rprf_R_t {
    RPRF_R_NONE,
    RPRF_R_PER_OUTER,
    RPRF_R_PER_INNER,
    RPRF_R_RATE
};

extern enum rprf_R_t rprf_R;

/* Types of summarized time reporting. */
enum rprf_S_t {
    RPRF_S_NONE,
    RPRF_S_MAX_RATE,
    RPRF_S_MIN_PER_INNER
};

extern enum rprf_S_t rprf_S;

extern int rprf_verbose;
extern int rprf_i_auto;
extern double rprf_i_min_t;
extern long rprf_i_min_m;
extern int rprf_t_comp;
extern double rprf_t_overhead;

extern void
rprf_parse_args(int argc, char* argv[], const char* format, ...);

/* Functions to allocate vectors from memory arena. */
#define rprf_malloc_vi8(len) \
    ((char *) rprf_malloc(len, sizeof(char)))

#define rprf_malloc_vi16(len) \
    ((short *) rprf_malloc((len)*sizeof(short), sizeof(short)))

#define rprf_malloc_vi32(len) \
    ((long *) rprf_malloc((len)*sizeof(long), sizeof(long)))

#define rprf_malloc_vf(len) \
    ((float *) rprf_malloc((len)*sizeof(float), sizeof(float)))

#define rprf_malloc_vd(len) \
    ((double *) rprf_malloc((len)*sizeof(double), sizeof(double)))

extern void *rprf_malloc(size_t len, size_t alignment);

#define RPRF_BODY(one_run, ops_arg) \
{                                                                            \
    double rprf_ops = (double) (ops_arg);                                    \
    IPM_timer rprf_t;                                                        \
                                                                             \
    long rprf_i;                                                             \
    double rprf_min_time;                                                    \
                                                                             \
    if (rprf_i_auto)                                                         \
    {                                                                        \
        double rprf_time = 0.0;                                              \
        double rprf_res = IPM_timer_resolution();                            \
        double rprf_min_res = rprf_res * rprf_i_min_m;                       \
                                                                             \
        rprf_min_time                                                        \
            = rprf_min_res > rprf_i_min_t ? rprf_min_res : rprf_i_min_t;     \
        rprf_min_time += rprf_t_overhead;                                    \
        if (rprf_verbose)                                                    \
        {                                                                    \
            printf("Starting automatic setting of inner loop count.\n");     \
            printf("Timer resolution reported as %gs.\n", rprf_res);         \
            printf("Requiring that inner loop take at least %gs.\n",         \
                   rprf_min_time);                                           \
        }                                                                    \
                                                                             \
        rprf_n_inner = 1;                                                    \
                                                                             \
        while (rprf_time < rprf_min_time)                                    \
        {                                                                    \
            long rprf_j;                                                     \
            const long rprf_n = rprf_n_inner;                                \
            double rprf_i_time = 0.0;                                        \
                                                                             \
            if (rprf_verbose)                                                \
            {                                                                \
                printf("Timing %ld inner loop(s)...\n", rprf_n_inner);       \
            }                                                                \
                                                                             \
            for (rprf_i = 0; rprf_i < rprf_n_outer; rprf_i++)                \
            {                                                                \
                IPM_timer_clear(&rprf_t);                                    \
                IPM_timer_start(&rprf_t);                                    \
                for (rprf_j=0; rprf_j<rprf_n; rprf_j++)                      \
                {                                                            \
                    one_run;                                                 \
                }                                                            \
                IPM_timer_stop(&rprf_t);                                     \
                rprf_time = IPM_timer_read(&rprf_t);                         \
                if (rprf_verbose)                                            \
                {                                                            \
                    printf("  %gs.\n", rprf_time);                           \
                }                                                            \
                if (rprf_i == 0 || rprf_time < rprf_i_time)                  \
                    rprf_i_time = rprf_time;                                 \
            }                                                                \
            if (rprf_i_time == 0.0)                                          \
            {                                                                \
                rprf_n_inner *= 10;                                          \
            }                                                                \
            else if (rprf_i_time < rprf_min_time)                            \
            {                                                                \
                const long rprf_n_inner_new = (long)                         \
                    ((1.1*(rprf_min_time/rprf_i_time)*rprf_n_inner)+0.51);   \
                if (rprf_n_inner_new <= rprf_n_inner)                        \
                    rprf_n_inner++;                                          \
                else                                                         \
                    rprf_n_inner = rprf_n_inner_new;                         \
            }                                                                \
        }                                                                    \
    }                                                                        \
    if (rprf_verbose)                                                        \
    {                                                                        \
        printf("Using inner loop count = %ld.\n", rprf_n_inner);             \
    }                                                                        \
                                                                             \
    rprf_min_time = 1e30;                                                    \
    for (rprf_i=0; rprf_i<rprf_n_outer; rprf_i++)                            \
    {                                                                        \
        long rprf_j;                                                         \
        const long rprf_n = rprf_n_inner;                                    \
        double rprf_time;                                                    \
                                                                             \
        IPM_timer_clear(&rprf_t);                                            \
        IPM_timer_start(&rprf_t);                                            \
        for (rprf_j=0; rprf_j<rprf_n; rprf_j++)                              \
        {                                                                    \
            one_run;                                                         \
        }                                                                    \
        IPM_timer_stop(&rprf_t);                                             \
        rprf_time = IPM_timer_read(&rprf_t);                                 \
        if (rprf_t_comp && rprf_time > rprf_t_overhead)                      \
            rprf_time -= rprf_t_overhead;                                    \
                                                                             \
        if (rprf_R!=RPRF_R_NONE || rprf_verbose)                             \
        {                                                                    \
            switch (rprf_R)                                                  \
            {                                                                \
            case RPRF_R_NONE:                                                \
            case RPRF_R_PER_OUTER:                                           \
                printf(IPM_timer_format() , rprf_time);                      \
                putchar(' ');                                                \
                break;                                                       \
            case RPRF_R_PER_INNER:                                           \
                printf("%g ", rprf_time/rprf_n);                             \
                break;                                                       \
            case RPRF_R_RATE:                                                \
                printf("%g ", (1e-6*(rprf_ops)*rprf_n)/rprf_time);           \
                break;                                                       \
            default:                                                         \
                assert(0);                                                   \
                break;                                                       \
            }                                                                \
        }                                                                    \
                                                                             \
        if (rprf_S!=RPRF_S_NONE)                                             \
        {                                                                    \
            if (rprf_time < rprf_min_time)                                   \
                rprf_min_time = rprf_time;                                   \
        }                                                                    \
    }                                                                        \
    if (rprf_R!=RPRF_R_NONE || rprf_verbose)                                 \
    {                                                                        \
        putchar('\n');                                                       \
    }                                                                        \
                                                                             \
    if (rprf_S!=RPRF_S_NONE)                                                 \
    {                                                                        \
        switch (rprf_S)                                                      \
        {                                                                    \
        case RPRF_S_MAX_RATE:                                                \
            printf("%8.6f\n", (1e-6*(rprf_ops)*rprf_n_inner)/rprf_min_time); \
            break;                                                           \
        case RPRF_S_MIN_PER_INNER:                                           \
            printf("%g\n", rprf_min_time/rprf_n_inner);                      \
            break;                                                           \
        default:                                                             \
            assert(0);                                                       \
            break;                                                           \
        }                                                                    \
    }                                                                        \
}

#ifdef __cplusplus
}
#endif

#endif
